//md2.cpp

#pragma comment(lib, "kernel32.lib")
#pragma comment(lib, "user32.lib")
#pragma comment(lib, "gdi32.lib")
#pragma comment(lib, "comctl32.lib")
#pragma comment(lib, "Advapi32.lib")
#pragma comment(lib, "Winmm.lib")

#include <windows.h>
#include <stdio.h>

#define ULONG_PTR unsigned long

#pragma comment(lib, "D3d9.lib") 
#pragma comment(lib, "D3dx9.lib")
#include <d3dx9.h>

#include "md2.h"


inline float Linear(float v0, float v1, float t)
{
    //return v0 * (1.0f - t) + v1 * t;
    // 
    return v0 + t * (v1 - v0); 

}


CMd2::CMd2()
{
	m_pFrames = NULL;
	m_pSkins = NULL;
	m_pTextureCoords = NULL;
	m_pTriIndex = NULL;

	m_aniKey.numKeys = 0;
	m_aniKey.pKey = NULL;

	m_vb = NULL;
	m_pTexture = NULL;

	m_iCurFrame = 0;
	m_lastTime = 0.0f;

	m_iAnimationKey = 0;
}
//-----------------------------------------------------------------------------------------------------
CMd2::~CMd2()
{
	//Release();
}
//-----------------------------------------------------------------------------------------------------
bool CMd2::Load(char* szFileName, char* szTextureName, LPDIRECT3DDEVICE9 pD3DDevice)
{
	FILE* fp = fopen(szFileName, "rb");
	if(fp == NULL)
		return false;

	ReadHeader(fp);
	ReadData(fp);

	fclose(fp);

	MakeDX(pD3DDevice);
	BuildAnimationKey();
	
	D3DXCreateTextureFromFile(pD3DDevice, szTextureName, &m_pTexture);
	return true;
}
//-----------------------------------------------------------------------------------------------------
bool CMd2::ReadHeader(FILE* fp)
{
	fread(&m_md2Header, 1, sizeof(m_md2Header), fp);	
	return true;
}
//-----------------------------------------------------------------------------------------------------
bool CMd2::ReadData(FILE* fp)
{
	m_pSkins         = new Md2Skins[ m_md2Header.numSkins ];
	m_pTextureCoords = new Md2TextureCoords[ m_md2Header.numTexCoords ];
	m_pTriIndex      = new Md2TriIndex[ m_md2Header.numTris ];
	m_pFrames		 = new Md2Frames[ m_md2Header.numFrames ];

	//ؽ  б
	fseek(fp, m_md2Header.offsetSkins, SEEK_SET);
	fread(m_pSkins, sizeof(Md2Skins), m_md2Header.numSkins, fp);

	//ؽ ǥ б
	fseek(fp, m_md2Header.offsetTexCoords, SEEK_SET);
	fread(m_pTextureCoords, sizeof(Md2TextureCoords), m_md2Header.numTexCoords, fp);

	//ﰢ  ε б
	fseek(fp, m_md2Header.offsetTris, SEEK_SET);
	fread(m_pTriIndex, sizeof(Md2TriIndex), m_md2Header.numTris, fp);

	fseek(fp, m_md2Header.offsetFrames, SEEK_SET);

	unsigned char* pTemp = new unsigned char [m_md2Header.frameSize + 100];
	TempMd2DataFrame* pTempFrame = (TempMd2DataFrame*) pTemp;

	for(int iFrame=0; iFrame< m_md2Header.numFrames; iFrame++)
	{
		// Ÿ о ȯѴ.
		fread(pTempFrame, 1, m_md2Header.frameSize, fp); 

		m_pFrames[iFrame].pData = new Md2Vertex[ m_md2Header.numVertices ];	
		strcpy( m_pFrames[iFrame].name,  pTempFrame->name );
		// ǥ ȭ Z<->Y
		for(int i=0; i< m_md2Header.numVertices; i++)
		{
			TempMd2DataVertex* pData = &(pTempFrame->data[i]);
			m_pFrames[iFrame].pData[i].vertex[0] = pData->vertex[0] * pTempFrame->scale[0] + pTempFrame->translate[0];      // x
			m_pFrames[iFrame].pData[i].vertex[2] = -1*(pData->vertex[1] * pTempFrame->scale[1] + pTempFrame->translate[1]); // z
			m_pFrames[iFrame].pData[i].vertex[1] = pData->vertex[2] * pTempFrame->scale[2] + pTempFrame->translate[2];      // y
		}
	}

	delete [] pTemp;

	return true;
}
//-------------------------------------------------------------------------------------------------------------------
void CMd2::BuildAnimationKey()
{
	char strName[16] = {0};
	char strLastName[16] = {0};
	char strFirstName[16] = {0};

	for(int iFrame=0; iFrame< m_md2Header.numFrames; iFrame++)
	{
		strcpy(strName, m_pFrames[iFrame].name);
		int stringLength = (int)strlen( strName );
		for(int i=0; i< stringLength; i++)
		{
			if( (strName[i] < 'A') || (strName[i] > 'z' ) )
			{
				strName[i] = '\0';
				if( strcmp( strLastName, strName ) )
				{
					strcpy(strLastName, strName);
					m_aniKey.numKeys++;

				    if( !strcmp( strFirstName, "" ) )
						strcpy( strFirstName, strName );
				}

				break;
			}
		}	
	}

	m_aniKey.pKey = new aniKeyFrame[ m_aniKey.numKeys ];

	strcpy( strName, "");
	strcpy( strLastName, strFirstName);
	int curKey = 0;

	m_aniKey.pKey[0].start = 0;
	strcpy( m_aniKey.pKey[0].szName, strFirstName );

	for(iFrame=0; iFrame< m_md2Header.numFrames; iFrame++)
	{
		strcpy(strName, m_pFrames[iFrame].name);
		int stringLength = (int)strlen( strName );
		for(int i=0; i< stringLength; i++)
		{
			if( (strName[i] < 'A') || (strName[i] > 'z' ) )
			{
				strName[i] = '\0';
				if( strcmp( strLastName, strName ) )
				{
					m_aniKey.pKey[curKey].end = iFrame-1;
					curKey++;
					m_aniKey.pKey[curKey].start = iFrame;

					strcpy(strLastName, strName);
					strcpy(m_aniKey.pKey[curKey].szName, strName);
#ifdef _DEBUG
					char szDebug[MAX_PATH];
					sprintf(szDebug, "%s\n", strName);
					OutputDebugString(szDebug);
#endif
				}
				else
				{
					m_aniKey.pKey[curKey].end = iFrame;
				}
				break;
			}
		}	
	}
}
//-------------------------------------------------------------------------------------------------------------------
void CMd2::MakeDX(LPDIRECT3DDEVICE9 pD3DDevice)
{
	if(pD3DDevice)
	{
		int numFaces = m_md2Header.numTris;

		//   ۸ Ѵ.
		pD3DDevice->CreateVertexBuffer( numFaces * 3 * sizeof(MyVertex), 0, MYVERETX_FVF, D3DPOOL_MANAGED, &m_vb, NULL );
	}
}
//-------------------------------------------------------------------------------------------------------------------
void CMd2::SetAnimationKey(int iAnimationKey)
{
	if(iAnimationKey == -1) 
		m_iAnimationKey++;

	if(m_iAnimationKey >= m_aniKey.numKeys)
		m_iAnimationKey = 0;

	m_iCurFrame = m_aniKey.pKey[ m_iAnimationKey ].start;
}
//-------------------------------------------------------------------------------------------------------------------
void CMd2::Update(float fTime)
{
	float  delay = 0.2f;
	float elapsedTime = fTime - m_lastTime;
	elapsedTime = elapsedTime / 1000.0f; //ʴ ȯ
	elapsedTime = elapsedTime / delay;   //delay ð  üũ

	if(elapsedTime >= 1.0f)
	{	
		m_lastTime = fTime;
		m_iCurFrame++;
		elapsedTime = 0.0f;
	}
	else
	{
		if(fTime - m_lastTime < 0.03)  //ʴ 30̶ 
			return;
	}

	if( m_iCurFrame > m_aniKey.pKey[ m_iAnimationKey ].end)
	    m_iCurFrame = m_aniKey.pKey[ m_iAnimationKey ].start;

	MakeDXVertices(m_iCurFrame, elapsedTime);
}
//-------------------------------------------------------------------------------------------------------------------
void CMd2::MakeDXVertices(int iFrame, float fLinearTime)
{
	MyVertex* pVertices = 0;
	m_vb->Lock(0,0, (void**)&pVertices, 0);
	int numFaces = m_md2Header.numTris;

	Md2Vertex*  pCurVertex  =  m_pFrames[iFrame].pData;

	//    ޽ ã´.
	int iNextFrame = m_iCurFrame + 1;
	if( iNextFrame > m_aniKey.pKey[ m_iAnimationKey ].end)
	    iNextFrame = m_aniKey.pKey[ m_iAnimationKey ].start;

	Md2Vertex*  pNextVertex =  m_pFrames[iNextFrame].pData;

	for(int i = 0; i < numFaces; ++i)
	{
		for(int v = 0; v < 3; ++v)
		{
			int index = m_pTriIndex[i].vertexIndex[v];
			float x1 = pCurVertex[index].vertex[0];//x
			float y1 = pCurVertex[index].vertex[1];//y
			float z1 = pCurVertex[index].vertex[2];//z

			float x2 = pNextVertex[index].vertex[0];//x
			float y2 = pNextVertex[index].vertex[1];//y
			float z2 = pNextVertex[index].vertex[2];//z

#if  0
			pVertices[i*3+v].x = x1;
			pVertices[i*3+v].y = y1;
			pVertices[i*3+v].z = z1;
#else
			pVertices[i*3+v].x = Linear(x1, x2, fLinearTime);;
			pVertices[i*3+v].y = Linear(y1, y2, fLinearTime);
			pVertices[i*3+v].z = Linear(z1, z2, fLinearTime);
#endif
			int texIndex = m_pTriIndex[i].textureIndex[v];
			pVertices[i*3+v].tu = m_pTextureCoords[ texIndex ].u / (float)m_md2Header.skinWidth;
			pVertices[i*3+v].tv = m_pTextureCoords[ texIndex ].v / (float)m_md2Header.skinHeight;
		}
	}
	m_vb->Unlock();
}
//-------------------------------------------------------------------------------------------------------------------
void CMd2::Render(LPDIRECT3DDEVICE9 pD3DDevice)
{
	unsigned int numFaces = m_md2Header.numTris;	

    pD3DDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
    pD3DDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
	pD3DDevice->SetTextureStageState(0,D3DTSS_COLORARG1, D3DTA_TEXTURE);
	pD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
	

	// Draw our triangle.
	pD3DDevice->SetStreamSource(0, m_vb, 0, sizeof(MyVertex));
	pD3DDevice->SetFVF(MYVERETX_FVF);
	pD3DDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, numFaces);
	pD3DDevice->SetTexture(0, m_pTexture);
}
//-------------------------------------------------------------------------------------------------------------------
void CMd2::Release()
{
	if(m_pTexture)
		m_pTexture->Release();
	if(m_vb)
		m_vb->Release();

	if(m_pFrames)
	{
		for(int iFrame=0; iFrame< m_md2Header.numFrames; iFrame++)
		{
			delete [] m_pFrames[iFrame].pData;
		}

		delete []  m_pFrames;
	}

	if(m_aniKey.pKey)
	{
		delete [] m_aniKey.pKey;
	}

	if(m_pSkins)
		delete [] m_pSkins;
	if(m_pTextureCoords)
		delete [] m_pTextureCoords;
	if(m_pTriIndex)
		delete [] m_pTriIndex;
}